# The MIT License (MIT)
#
# Copyright (c) 2018 Niklas Rosenstein
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.
"""
Provides simple sumtype declarations.
"""

from six.moves import zip
from six import iteritems, with_metaclass

import re
import types
import six

__all__ = ['Constructor', 'MemberOf', 'Sumtype', 'AddIsMethods']


class Constructor(object):
  """
  Represents a constructor for a sumtype.
  """

  def __init__(self, *args):
    self.name = None
    self.type = None
    self.args = args
    self.members = {}

  def bind(self, type, name):
    obj = Constructor(*self.args)
    obj.type = type
    obj.name = name
    obj.members = self.members.copy()
    return obj

  def accept_args(self, *args):
    if self.type is None or self.name is None:
      raise RuntimeError('unbound Constructor')
    if len(args) != len(self.args):
      raise TypeError('{}.{}() expected {} arguments, got {}'.format(
        self.type.__name__, self.name, len(self.args), len(args)))
    return dict(zip(self.args, args))

  def __call__(self, *args):
    if self.type is None or self.name is None:
      raise RuntimeError('unbound Constructor')
    return self.type(self, self.accept_args(*args))


class MemberOf(object):
  """
  A decorator for functions or values that are supposed to be members of
  only a specific sumtype's constructor (or multiple constructors). Instances
  of this class will be automatically unpacked by the :class:`_TypeMeta`
  constructor and moved into the :attr:`Constructor.members` dictionary.
  """

  def __init__(self, constructors=None, value=None, name=None):
    if isinstance(constructors, Constructor):
      constructors = [constructors]
    self.constructors = constructors
    self.value = value
    self.name = name

    if name:
      for c in constructors:
        c.members[name] = value

  def __call__(self, value):
    if not self.name:
      self.name = value.__name__
    for c in self.constructors:
      c.members[self.name] = value
    return self

  def update_constructors(self, attrname):
    for constructor in self.constructors:
      constructor.members[attrname] = self.value


class _TypeMeta(type):

  def __new__(cls, name, bases, attrs):
    subtype = type.__new__(cls, name, bases, attrs)

    # Collect all new constructors.
    constructors = getattr(subtype, '__constructors__', {}).copy()
    for key, value in iteritems(vars(subtype)):
      if isinstance(value, Constructor):
        constructors[key] = value

    # Update constructors from MemberOf declarations.
    for key, value in list(iteritems(vars(subtype))):
      if isinstance(value, MemberOf):
        delattr(subtype, key)

    # Bind constructors.
    for key, value in iteritems(constructors):
      setattr(subtype, key, value.bind(subtype, key))
    subtype.__constructors__ = constructors

    # Invoke addons.
    for addin in getattr(subtype, '__addins__', []):
      addin.init_type(subtype)

    return subtype

  def __getattr__(self, name):
    if self.__bases__ == (object,) and name in __all__:
      return globals()[name]
    raise AttributeError(name)


class Sumtype(with_metaclass(_TypeMeta)):
  """
  Base class for sumtypes. You can access all members of the sumtype
  module via this type (but not through subclasses).

  ```python
  from nr.types import Sumtype

  class Result(Sumtype):
    Ok = Sumtype.Constructor()
    Error = Sumtype.Constructor('message')

  assert not hasattr(Result, 'Constructor')
  ```
  """

  __addins__ = []
  __constructors__ = {}

  def __init__(self, constructor, attrs):
    assert isinstance(constructor, Constructor), type(constructor)
    assert isinstance(attrs, dict), type(attrs)
    self.__constructor__ = constructor
    for key, value in iteritems(constructor.members):
      if isinstance(value, types.FunctionType):
        value = value.__get__(self, constructor.type)
      setattr(self, key, value)
    for key in constructor.args:
      if key not in attrs:
        raise ValueError('missing key in attrs: {!r}'.format(key))
    for key, value in iteritems(attrs):
      if key not in constructor.args:
        raise ValueError('unexpected key in attrs: {!r}'.format(key))
      setattr(self, key, value)

  def __getitem__(self, index):
    if hasattr(index, '__index__'):
      index = index.__index__()
    if isinstance(index, int):
      return getattr(self, self.__constructor__.args[index])
    elif isinstance(index, slice):
      return tuple(getattr(self, k) for k in self.__constructor__.args[index])
    else:
      raise TypeError('indices must be integers or slices, not str')

  def __iter__(self):
    for k in self.__constructor__.args:
      yield getattr(self, k)

  def __len__(self):
    return len(self.__constructor__.args)

  def __repr__(self):
    return '{}.{}({})'.format(type(self).__name__, self.__constructor__.name,
      ', '.join('{}={!r}'.format(k, getattr(self, k)) for k in self.__constructor__.args))


class AddIsMethods(object):

  @staticmethod
  def init_type(type):
    def create_is_check(func_name, constructor_name):
      def check(self):
        constructor = getattr(self, constructor_name)
        return self.__constructor__ == constructor
      check.__name__ = name
      check.__qualname__ = name
      return check
    for name in type.__constructors__.keys():
      func_name = 'is_' + '_'.join(re.findall('[A-Z]+[^A-Z]*', name)).lower()
      setattr(type, func_name, create_is_check(func_name, name))


Sumtype.__addins__.append(AddIsMethods)



# The MIT License (MIT)
#
# Copyright (c) 2018 Niklas Rosenstein
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.

from .map import OrderedDict
import six


class _NamedMeta(type):
  """
  Metaclass for the #Named class.
  """

  def __init__(self, name, bases, data):
    # Inherit the annotations of the base classes, in the correct order.
    annotations = getattr(self, '__annotations__', {})
    if isinstance(annotations, (list, tuple)):
      for i, item in enumerate(annotations):
        if len(item) == 3:
          setattr(self, item[0], item[2])
          item = item[:2]
          annotations[i] = item
      annotations = OrderedDict(annotations)
    new_annotations = OrderedDict()
    for base in bases:
      base_annotations = getattr(base, '__annotations__', {})
      if isinstance(base_annotations, (list, tuple)):
        base_annotations = OrderedDict(base_annotations)
      for key, value in base_annotations.items():
        if key not in annotations:
          new_annotations[key] = value
    new_annotations.update(annotations)
    self.__annotations__ = new_annotations
    super(_NamedMeta, self).__init__(name, bases, data)

  def __getattr__(self, name):
    if self.__bases__ == (object,) and name == 'Initializer':
      return Initializer
    raise AttributeError(name)


class Named(six.with_metaclass(_NamedMeta)):
  """
  A base-class similar to #typing.NamedTuple, but mutable. Fields can be
  specified using Python 3.6 class-member annotations or by setting the
  `__annotations__` field to a list where each item is a member declaration
  that consists of two or three items where 1) is the name, 2) is the
  annotated value (type) and 3) is the default value for the field.

  Note that you can access the #Initializer class via this class, but not
  through subclasses.

  ```python
  from nr.types import Named

  class Person(Named):
    __annotations__ = [
      ('mail', str),
      ('name', str, Named.Initializer(random_name)),
      ('age', int, 0)
    ]

  class Person(Named):
    # Python 3.6+
    mail: str
    name: str = Named.Initializer(random_name)
    age: int = 0
  ```
  """

  def __init__(self, *args, **kwargs):
    annotations = getattr(self, '__annotations__', {})
    if len(args) > len(annotations):
      raise TypeError('{}() expected {} positional arguments, got {}'
        .format(type(self).__name__, len(annotations), len(args)))
    if isinstance(annotations, (list, tuple)):
      annotations = OrderedDict(annoations)

    for arg, (key, ant) in zip(args, annotations.items()):
      setattr(self, key, arg)
      if key in kwargs:
        raise TypeError('{}() duplicate value for argument "{}"'
          .format(type(self).__name__, key))

    for key, ant in annotations.items():
      if key in kwargs:
        setattr(self, key, kwargs.pop(key))
      else:
        try:
          value = getattr(self, key)
        except AttributeError:
          raise TypeError('{}() missing argument "{}"'
            .format(type(self).__name__, key))
        if isinstance(value, Initializer):
          setattr(self, key, value.func())

    for key in kwargs.keys():
      raise TypeError('{}() unexpected keyword argument "{}"'
        .format(type(self).__name__, key))

  def __repr__(self):
    members = ', '.join('{}={!r}'.format(k, getattr(self, k)) for k in self.__annotations__)
    return '{}({})'.format(type(self).__name__, members)

  def __iter__(self):
    for key in self.__annotations__:
      yield getattr(self, key)

  def asdict(self):
    return {k: getattr(self, k) for k in self.__annotations__}


class Initializer:
  """
  Use this for the default value of annotated fields to wrap a function that
  will be called to retrieve the default value for the field. Works only with
  #Named as the base class.
  """

  def __init__(self, func):
    self.func = func




__author__ = 'Niklas Rosenstein <rosensteinniklas@gmail.com>'
__version__ = '1.0.0'

from .enum import Enumeration
from .map import ChainMap, MapAsObject, ObjectAsMap, OrderedDict
from .named import Named
from .record import Record
from .sumtype import Sumtype
from .meta import InlineMetaclassBase



# The MIT License (MIT)
#
# Copyright (c) 2018 Niklas Rosenstein
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.
"""
This module implements a simple framework for creating C-like enumerations in
Python using classes. Simply inherit from the #Enumeration class.

```python
>>> class Color(enum.Enumeration):
...   red = 0
...   green = 1
...   blue = 2
>>> print Color.red
<Color: red>
>>> print Color('green')
<Color: green>
>>> print Color(2)
<Color: blue>
>>> print Color('red') is Color.red
True
>>> print Color.blue.name
blue
>>> Color(343)
Traceback (most recent call last):
  File "test.py", line 10, in <module>
  Color(343)
  File "C:\\repositories\\py-nr.utils\\nr\\utils\\enum.py", line 159, in __new__
  raise NoSuchEnumerationValue(cls.__name__, value)
nr.utils.enum.NoSuchEnumerationValue: ('Color', 343)
```

If you want to disable that an invalid enumeration value will raise an error,
a `__fallback__` value can be specified on class-level.

```python
>>> class Color(enum.Enumeration):
...   red = 0
...   green = 1
...   blue = 2
...   __fallback__ = -1
>>> print Color(42)
<Color -invalid->
>>> print Color(7).value
-1
>>> print Color(16).name
-invalid-
```

You can also iterate over an enumeration class. Note that the order of the
items yielded is value-sorted and the order of declaration does not play any
role.

```python
>>> class Color(enum.Enumeration):
...   red = 0
...   green = 1
...   blue = 2
...   __fallback__ = -1
>>> for color in Color:
...   print color
<Color: red>
<Color: green>
<Color: blue>
```

You can add data or actual methods to an enumeration class by wrapping it with
the #Data class.

```python
class Color(enum.Enumeration):
  red = 0
  green = 1
  blue = 2

  @enum.Data
  @property
  def astuple(self):
    if self == Color.red:
      return (1, 0, 0)
    elif self == Color.green:
      return (0, 1, 0)
    elif self == Color.blue:
      return (0, 0, 1)
    else:
      assert False

print Color.red.astuple
# (1, 0, 0)
```
"""

import ctypes
import six
import sys


class NoSuchEnumerationValue(ValueError):
  r""" Raised when an Enumeration object was attempted to be
  created from an integer value but there was no enumeration
  object for this value.

  Note that you can specify ``__fallback_value__`` on an
  Enumeration class to not let it raise an exception. """

  pass


class Data(object):
  """
  Small class that can be used to specify data on an enumeration that should
  not be converted and interpreted as an enumeration value.

  ```python
  class Color(enum.Enumeration):
    red = 0
    green = 1
    blue = 2

    @enum.Data
    @property
    def astuple(self):
      if self == Color.red:
        return (1, 0, 0)
      elif self == Color.green:
        return (0, 1, 0)
      elif self == Color.blue:
        return (0, 0, 1)
      else:
        assert False

  print Color.red.astuple
  # (1, 0, 0)
  ```

  This class can be subclassed to add new sugar to the already very sweet pie.
  """

  def __init__(self, value):
    super(Data, self).__init__()
    self.value = value

  def unpack(self):
    return self.value


class _EnumerationMeta(type):
  """
  This is the meta class for the #Enumeration base class which handles the
  automatic conversion of integer values to instances of the #Enumeration
  class. There are no other types allowed other than int or #Data which
  will be unpacked on the Enumeration class.

  If a `__fallback__` was defined on class-level as an integer, the
  #Enumeration constructor will not raise a #NoSuchEnumerationValue exception
  if the passed value did not match the enumeration values, but instead return
  that fallback value.

  This fallback is not taken into account when attempting to create a new
  #Enumeration object by a string.
  """

  _values = None
  __fallback__ = None

  def __new__(cls, name, bases, data):

    # Unpack all Data objects and create a dictionary of
    # values that will be converted to instances of the
    # enumeration class later.
    enum_values = {}
    collections = {}
    for key, value in data.items():
      # Unpack Data objects into the class.
      if isinstance(value, Data):
        data[key] = value.unpack()

      # Integers will be enumeration values.
      elif isinstance(value, int):
        enum_values[key] = value

      # Lists and tuples will be converted to
      # collections of the Enumeration values.
      elif isinstance(value, (list, tuple)):
        collections[key] = value

      # We don't accept anything else.
      elif not key.startswith('_'):
        message = 'Enumeration must consist of ints or Data objects ' \
              'only, got %s for \'%s\''
        raise TypeError(message % (value.__class__.__name__, key))

    # Create the new class object and give it the dictionary
    # that will map the integral values to the instances.
    class_ = type.__new__(cls, name, bases, data)
    class_._values = {}

    # Iterate over all entries in the data entries and
    # convert integral values to instances of the enumeration
    # class.
    for key, value in six.iteritems(enum_values):

      # Create the new object. We must not use the classes'
      # __new__() method as it resolves the object from the
      # existing values.
      obj = object.__new__(class_)
      object.__init__(obj)

      obj.value = value
      obj.name = key

      if key == '__fallback__':
        obj.name = '-invalid-'
      else:
        class_._values[value] = obj
      setattr(class_, key, obj)


    # Convert the collections.
    for key, value in six.iteritems(collections):
      value = type(value)(class_(v) for v in value)
      setattr(class_, key, value)

    return class_

  def __iter__(self):
    " Iterator over value-sorted enumeration values. "

    return iter(self.__values__())

  def __values__(self):
    return sorted(six.itervalues(self._values), key=lambda x: x.value)

  def __getattr__(self, name):
    if self.__bases__ == (object,) and name == 'Data':
      return Data
    raise AttributeError(name)


class Enumeration(six.with_metaclass(_EnumerationMeta)):
  """
  This is the base class for listing enumerations. All components of the class
  that are integers will be automatically converted to instances of the
  #Enumeration class. Creating new instances of the class will only work if the
  value is an existing enumeration value.

  The hash of an enumeration value is its name, but indexing a container
  corresponds to its value.
  """

  def __new__(cls, value, _allow_fallback=True):
    """
    Creates a new instance of the Enumeration. *value* must be the integral
    number of one of the existing enumerations. #NoSuchEnumerationValue is
    raised in any other case.

    If a fallback was defined, it is returned only if *value* is an integer,
    not if it is a string.
    """

    # Try to find the actual instance of the Enumeration class
    # for the integer value and return it if it is available.
    if isinstance(value, six.integer_types):
      try:
        value = cls._values[value]
      except KeyError:

        # If a fallback value was specified, use it
        # instead of raising an exception.
        if _allow_fallback and cls.__fallback__ is not None:
          return cls.__fallback__

        raise NoSuchEnumerationValue(cls.__name__, value)

    # Or by name?
    elif isinstance(value, six.string_types):
      try:
        new_value = getattr(cls, value)
        if type(new_value) != cls:
          raise AttributeError
      except AttributeError:
        raise NoSuchEnumerationValue(cls.__name__, value)

      value = new_value

    # At this point, value must be an object of the Enumeration
    # class, otherwise an invalid value was passed.
    if type(value) == cls:
      return value

    raise TypeError('value must be %s or int, got %s' % (cls.__name__, type(value).__name__))

  def __hash__(self):
    return hash(self.name)

  def __eq__(self, other):
    if type(other) == self.__class__:
      return other.value == self.value
    elif isinstance(other, six.string_types):
      return other == self.name
    else:
      return self.value == other

  def __ne__(self, other):
    return not (self == other)

  def __int__(self):
    return self.value

  def __str__(self):
    class_name = self.__class__.__name__
    return '<%s: %s>' % (class_name, self.name)

  def __repr__(self):
    class_name = self.__class__.__name__
    return '<%s: [%d] %s>' % (class_name, self.value, self.name)

  def __index__(self):
    return self.value

  def __nonzero__(self):
    return False
  __bool__ = __nonzero__ # Py3

  # ctypes support

  @property
  def _as_parameter_(self):
    return ctypes.c_int(self.value)

  @Data
  @classmethod
  def from_param(cls, obj):
    if isinstance(obj, (int,) + six.string_types):
      obj = cls(obj)
    if type(obj) != cls:
      c1 = cls.__name__
      c2 = obj.__class__.__name__
      raise TypeError('can not create %s from %s' % (c1, c2))

    return ctypes.c_int(obj.value)




import setuptools

def readme():
  with open('README.md', encoding='utf8') as fp:
    return fp.read()

def requirements():
  with open('requirements.txt') as fp:
    return fp.readlines()

setuptools.setup(
  name = 'nr.types',
  version = '1.0.1',
  author = 'Niklas Rosenstein',
  author_email = 'rosensteinniklas@gmail.com',
  description = 'Better named tuples, sumtypes and useful map types.',
  long_description = readme(),
  long_description_content_type = 'text/markdown',
  url = 'https://github.com/NiklasRosenstein-Python/nr.types',
  license = 'MIT',
  packages = setuptools.find_packages('src'),
  package_dir = {'': 'src'}
)



# The MIT License (MIT)
#
# Copyright (c) 2018 Niklas Rosenstein
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.

import six

try:
  from collections import OrderedDict
except ImportError:
  from ._ordereddict import OrderedDict

if six.PY2:
  _can_iteritems = lambda x: hasattr(x, 'iteritems')
  _can_iterkeys = lambda x: hasattr(x, 'keys')
else:
  _can_iteritems = lambda x: hasattr(x, 'items')
  _can_iterkeys = lambda x: hasattr(x, 'keys')


class ObjectAsMap(object):
  """
  This class wraps an object and exposes its members as mapping.
  """

  def __new__(cls, obj):
    if isinstance(obj, MapAsObject):
      return obj._ObjectFromMapping__mapping
    return super(ObjectAsMap, cls).__new__(cls)

  def __init__(self, obj):
    self.__obj = obj

  def __repr__(self):
    return 'ObjectAsMap({!r})'.format(self.__obj)

  def __iter__(self):
    return self.keys()

  def __len__(self):
    return len(dir(self.__obj))

  def __contains__(self, key):
    return hasattr(self.__obj, key)

  def __getitem__(self, key):
    try:
      return getattr(self.__obj, key)
    except AttributeError:
      raise KeyError(key)

  def __setitem__(self, key, value):
    setattr(self.__obj, key, value)

  def __delitem__(self, key):
    delattr(self.__obj, key)

  def keys(self):
    return iter(dir(self.__obj))

  def values(self):
    return (getattr(self.__obj, k) for k in dir(self.__obj))

  def items(self):
    return ((k, getattr(self.__obj, k)) for k in dir(self.__obj))

  def get(self, key, default=None):
    return getattr(self.__obj, key, default)

  def setdefault(self, key, value):
    try:
      return getattr(self.__obj, key)
    except AttributeError:
      setattr(self.__obj, key, value)
      return value


class MapAsObject(object):
  """
  This class wraps a dictionary and exposes its values as members.
  """

  def __new__(cls, mapping, name=None):
    if isinstance(mapping, ObjectAsMap):
      return mapping._MappingFromObject__obj
    return super(MapAsObject, cls).__new__(cls)

  def __init__(self, mapping, name=None):
    self.__mapping = mapping
    self.__name = name

  def __getattribute__(self, key):
    if key.startswith('_MapAsObject__'):
      return super(MapAsObject, self).__getattribute__(key)
    try:
      return self.__mapping[key]
    except KeyError:
      raise AttributeError(key)

  def __setattr__(self, key, value):
    if key.startswith('_MapAsObject__'):
      super(MapAsObject, self).__setattr__(key, value)
    else:
      self.__mapping[key] = value

  def __delattr__(self, key):
    if key.startswith('_MapAsObject__'):
      super(MapAsObject, self).__delattr__(key)
    else:
      del self.__mapping[key]

  def __dir__(self):
    return sorted(self.__mapping.keys())

  def __repr__(self):
    if self.__name:
      return '<MapAsObject name={!r}>'.format(self.__name)
    else:
      return '<MapAsObject {!r}>'.format(self.__mapping)


class ChainMap(object):
  """
  A dictionary that wraps a list of dictionaries. The dictionaries passed
  into the #ChainMap will not be mutated. Setting and deleting values will
  happen on the first dictionary passed.
  """

  def __init__(self, *dicts):
    if not dicts:
      raise ValueError('need at least one argument')
    self._major = dicts[0]
    self._dicts = list(dicts)
    self._deleted = set()
    self._in_repr = False

  def __contains__(self, key):
    if key not in self._deleted:
      for d in self._dicts:
        if key in d:
          return True
    return False

  def __getitem__(self, key):
    if key not in self._deleted:
      for d in self._dicts:
        try: return d[key]
        except KeyError: pass
    raise KeyError(key)

  def __setitem__(self, key, value):
    self._major[key] = value
    self._deleted.discard(key)

  def __delitem__(self, key):
    if key not in self:
      raise KeyError(key)
    self._major.pop(key, None)
    self._deleted.add(key)

  def __iter__(self):
    return six.iterkeys(self)

  def __len__(self):
    return sum(1 for x in self.keys())

  def __repr__(self):
    if self._in_repr:
      return 'ChainMap(...)'
    else:
      self._in_repr = True
      try:
        return 'ChainMap({})'.format(dict(six.iteritems(self)))
      finally:
        self._in_repr = False

  def __eq__(self, other):
    return dict(self.items()) == other

  def __ne__(self, other):
    return not (self == other)

  def pop(self, key, default=NotImplemented):
    if key not in self:
      if default is NotImplemented:
        raise KeyError(key)
      return default
    self._major.pop(key, None)
    self._deleted.add(key)

  def popitem(self):
    if self._major:
      key, value = self._major.popitem()
      self._deleted.add(key)
      return key, value
    for d in self._dicts:
      for key in six.iterkeys(d):
        if key not in self._deleted:
          self._deleted.add(key)
          return key, d[key]
    raise KeyError('popitem(): dictionary is empty')

  def clear(self):
    self._major.clear()
    self._deleted.update(six.iterkeys(self))

  def copy(self):
    return type(self)(*self._dicts[1:])

  def setdefault(self, key, value=None):
    try:
      return self[key]
    except KeyError:
      self[key] = value
      return value

  def update(self, E, *F):
    if _can_iteritems(E):
      for k, v in six.iteritems(E):
        self[k] = v
    elif _can_iterkeys(E):
      for k in six.iterkeys(E):
        self[k] = E[k]
    elif hasattr(E, 'items'):
      E = E.items()
    for k, v in E:
      self[k] = v
    for Fv in F:
      for k, v in six.iteritems(Fv):
        self[k] = v

  def keys(self):
    seen = set()
    for d in self._dicts:
      for key in six.iterkeys(d):
        if key not in seen and key not in self._deleted:
          yield key
          seen.add(key)

  def values(self):
    seen = set()
    for d in self._dicts:
      for key, value in six.iteritems(d):
        if key not in seen and key not in self._deleted:
          yield value
          seen.add(key)

  def items(self):
    seen = set()
    for d in self._dicts:
      for key, value in six.iteritems(d):
        if key not in seen and key not in self._deleted:
          yield key, value
          seen.add(key)

  if six.PY2:
    iterkeys = keys
    keys = lambda self: list(self.iterkeys())
    itervalues = values
    values = lambda self: list(self.itervalues())
    iteritems = items
    items = lambda self: list(self.iteritems())



# The MIT License (MIT)
#
# Copyright (c) 2018 Niklas Rosenstein
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.

class Record(object):
  """
  This class can be considered a new base for all classes that implement
  the `__slots__` interface. It provides convenient methods to treat the
  class instances like mutable versions of #:collections.namedtuple.

  # Example
  ```python
  class MyRecord(Record):
    __slots__ = 'foo bar ham egg'.split()
    __defaults__ = {'egg': 'yummy'}
  # or
  MyRecord = Record.new('MyRecord', 'foo bar ham', egg='yummy')

  data = MyRecord('the-foo', 42, ham="spam")
  assert data.egg = 'yummy'
  ```
  """

  def __init__(self, *args, **kwargs):
    defaults = getattr(self, '__defaults__', None) or {}
    if len(args) > len(self.__slots__):
      msg = '__init__() takes {0} positional arguments but {1} were given'
      raise TypeError(msg.format(len(self.__slots__), len(args)))
    for key, arg in zip(self.__slots__, args):
      if key in kwargs:
        msg = 'multiple values for argument {0!r}'.format(key)
        raise TypeError(msg)
      kwargs[key] = arg
    for key, arg in kwargs.items():
      setattr(self, key, arg)
    for key in self.__slots__:
      if key not in kwargs:
        if key in defaults:
          setattr(self, key, defaults[key])
        else:
          raise TypeError('missing argument {0!r}'.format(key))
      else:
        kwargs.pop(key)
    if kwargs:
      msg = '__init__() got an unexpected keyword argument {0!r}'
      raise TypeError(msg.format(next(iter(kwargs))))

  def __repr__(self):
    parts = ['{0}={1!r}'.format(k, v) for k, v in self.items()]
    return '{0}('.format(type(self).__name__) + ', '.join(parts) + ')'

  def __iter__(self):
    """
    Iterate over the values of the record in order.
    """

    for key in self.__slots__:
      yield getattr(self, key)

  def __len__(self):
    """
    Returns the number of slots in the record.
    """

    return len(self.__slots__)

  def __getitem__(self, index_or_key):
    """
    Read the value of a slot by its index or name.

    :param index_or_key:
    :raise TypeError:
    :raise IndexError:
    :raise KeyError:
    """

    if isinstance(index_or_key, int):
      return getattr(self, self.__slots__[index_or_key])
    elif isinstance(index_or_key, str):
      if index_or_key not in self.__slots__:
        raise KeyError(index_or_key)
      return getattr(self, index_or_key)
    else:
      raise TypeError('expected int or str')

  def __setitem__(self, index_or_key, value):
    """
    Set the value of a slot by its index or name.

    :param index_or_key:
    :raise TypeError:
    :raise IndexError:
    :raise KeyError:
    """

    if isinstance(index_or_key, int):
      setattr(self, self.__slots__[index_or_key], value)
    elif isinstance(index_or_key, str):
      if index_or_key not in self.__slots__:
        raise KeyError(index_or_key)
      setattr(self, index_or_key, value)
    else:
      raise TypeError('expected int or str')

  def __eq__(self, other):
    for key in self.__slots__:
      try:
        other_value = getattr(other, key)
      except AttributeError:
        return False
      if getattr(self, key) != other_value:
        return False
    return True

  def __setattr__(self, name, value):
    if name in self.__slots__:
      setter = getattr(self, '_set_' + name, None)
      if callable(setter):
        setter(value)
        return
    super(Record, self).__setattr__(name, value)

  def __getattribute__(self, name):
    if name != '__slots__' and name in self.__slots__:
      getter = getattr(self, '_get_' + name, None)
      if callable(getter):
        return getter()
    return super(Record, self).__getattribute__(name)

  def items(self):
    """
    Iterator for the key-value pairs of the record.
    """

    for key in self.__slots__:
      yield key, getattr(self, key)

  def keys(self):
    """
    Iterator for the member names of the record.
    """

    return iter(self.__slots__)

  def values(self):
    """
    Iterator for the values of the object, like :meth:`__iter__`.
    """

    for key in self.__slots__:
      yield getattr(self, key)

  def _asdict(self):
    return dict((k, getattr(self, k)) for k in self.__slots__)

  @classmethod
  def new(cls, __name, __fields, **defaults):
    '''
    Creates a new class that can represent a record with the
    specified *fields*. This is equal to a mutable namedtuple.
    The returned class also supports keyword arguments in its
    constructor.

    :param __name: The name of the Record.
    :param __fields: A string or list of field names.
    :param defaults: Default values for fields. The defaults
      may list field names that haven't been listed in *fields*.
    '''

    name = __name
    fields = __fields
    fieldset = set(fields)

    if isinstance(fields, str):
      if ',' in fields:
        fields = fields.split(',')
      else:
        fields = fields.split()
    else:
      fields = list(fields)

    for key in defaults.keys():
      if key not in fields:
        fields.append(key)

    class _record(cls):
      __slots__ = fields
      __defaults__ = defaults
    _record.__name__ = name
    return _record
